import logging
import logging.config
import os
import typing as t
from importlib.metadata import version

from pythonjsonlogger.orjson import OrjsonFormatter

from .config import EnvironConfigFactory
from .trace import (
    new_sampled_trace,
    new_trace,
    new_trace_cm,
    notrace,
    setup_sentry,
    trace,
    trace_cm,
)

__version__ = version(__package__)

__all__ = [
    "init_logging",
    "AllowLessThanFilter",
    "notrace",
    "setup_sentry",
    "trace",
    "trace_cm",
    "new_sampled_trace",
    "new_trace",
    "new_trace_cm",
]


class AllowLessThanFilter(logging.Filter):
    def __init__(self, level: t.Union[int, str] = logging.ERROR, name: str = ""):
        super().__init__(name)
        if not isinstance(level, int):
            try:
                level = logging._nameToLevel[level]
            except KeyError:
                raise ValueError(f"Unknown level name: {level}")
        self.level = level

    def filter(self, record: logging.LogRecord) -> bool:
        return record.levelno < self.level


class _HealthCheckFilter(logging.Filter):
    def __init__(self, url_path: str = "/api/v1/ping", name: str = "") -> None:
        super().__init__(name)
        self.url_path = url_path

    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno > logging.INFO:
            return True
        return record.getMessage().find(self.url_path) == -1


BASE_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "standard": {"format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s"},
        "json": {
            "()": OrjsonFormatter,
            "reserved_attrs": [
                "created",
                "exc_text",
                "levelno",
                "msecs",
                "msg",
                "relativeCreated",
            ],
            "rename_fields": {
                "name": "logName",
                "levelname": "severity",
                "extra": "jsonPayload",
            },
            "timestamp": True,
        },
    },
    "filters": {
        "hide_errors": {"()": AllowLessThanFilter, "level": "ERROR"},
        "hide_health_checks": {"()": _HealthCheckFilter},
    },
    "handlers": {
        "stdout": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "standard",
            "stream": "ext://sys.stdout",
            "filters": ["hide_errors"],
        },
        "stderr": {
            "class": "logging.StreamHandler",
            "level": "ERROR",
            "formatter": "standard",
            "stream": "ext://sys.stderr",
        },
        "json": {
            "class": "logging.StreamHandler",
            "level": "DEBUG",
            "formatter": "json",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "aiohttp.access": {
            "level": logging.NOTSET,
            "propagate": True,
            "filters": ["hide_health_checks"],
        },
        "uvicorn.access": {
            "level": logging.NOTSET,
            "propagate": True,
            "filters": ["hide_health_checks"],
        },
    },
}


TEXT_CONFIG = BASE_CONFIG | {
    "root": {
        "level": logging.DEBUG,
        "handlers": ["stderr", "stdout"],
    },
}

JSON_CONFIG = BASE_CONFIG | {
    "root": {
        "level": logging.DEBUG,
        "handlers": ["json"],
    },
}


def init_logging(
    *,
    health_check_url_path: str = "/api/v1/ping",
) -> None:
    config = EnvironConfigFactory().create_logging()
    if "PYTEST_VERSION" in os.environ:
        config_template = TEXT_CONFIG
    else:
        config_template = JSON_CONFIG
    dict_config: dict[str, t.Any] = config_template.copy()
    dict_config["root"]["level"] = config.log_level
    if config.log_health_check:
        dict_config["loggers"].pop("aiohttp.access", None)
        dict_config["loggers"].pop("uvicorn.access", None)
    dict_config["filters"]["hide_health_checks"]["url_path"] = health_check_url_path
    logging.config.dictConfig(dict_config)



import logging

from neuro_logging.config import EnvironConfigFactory


class TestEnvironConfigFactory:
    def test_create_logging__defaults(self) -> None:
        config = EnvironConfigFactory({}).create_logging()

        assert config.log_level == logging.INFO
        assert not config.log_health_check

    def test_create_logging__custom(self) -> None:
        environ = {
            "LOG_LEVEL": "error",
            "LOG_HEALTH_CHECK": "1",
        }
        config = EnvironConfigFactory(environ).create_logging()

        assert config.log_level == logging.ERROR
        assert config.log_health_check

    def test_create_sentry__defaults(self) -> None:
        config = EnvironConfigFactory({}).create_sentry()

        assert config
        assert config.dsn is None
        assert config.cluster_name is None
        assert config.app_name is None
        assert config.sample_rate == 0.1

    def test_create_sentry__custom(self) -> None:
        environ = {
            "SENTRY_DSN": "sentry-dsn",
            "SENTRY_CLUSTER_NAME": "cluster",
            "SENTRY_APP_NAME": "app",
            "SENTRY_SAMPLE_RATE": "0.5",
        }
        config = EnvironConfigFactory(environ).create_sentry()

        assert config
        assert config.dsn == "sentry-dsn"
        assert config.cluster_name == "cluster"
        assert config.app_name == "app"
        assert config.sample_rate == 0.5



import asyncio
import functools
import inspect
import logging
from collections.abc import AsyncIterator, Awaitable, Callable, Iterable, Mapping
from contextlib import asynccontextmanager
from importlib.metadata import version
from typing import Any, Optional, TypeVar, Union, cast

import aiohttp
import sentry_sdk
from sentry_sdk import Hub
from sentry_sdk.integrations.aiohttp import AioHttpIntegration
from sentry_sdk.types import Event, Hint
from yarl import URL

from .config import EnvironConfigFactory

LOGGER = logging.getLogger(__name__)


T = TypeVar("T", bound=Callable[..., Awaitable[Any]])


@asynccontextmanager
async def new_sentry_trace_cm(
    name: str, sampled: bool
) -> AsyncIterator[sentry_sdk.tracing.Span]:
    with Hub(Hub.current) as hub:
        with hub.configure_scope() as scope:
            scope.clear_breadcrumbs()

        with hub.start_transaction(name=name, sampled=sampled) as transaction:
            try:
                yield transaction
            except asyncio.CancelledError:
                transaction.set_status("cancelled")
                raise
            except Exception as exc:
                hub.capture_exception(error=exc)
                raise


@asynccontextmanager
async def new_trace_cm(name: str, sampled: bool = False) -> AsyncIterator[None]:
    async with new_sentry_trace_cm(name, sampled):
        yield


@asynccontextmanager
async def sentry_trace_cm(
    name: str,
    tags: Optional[Mapping[str, str]] = None,
    data: Optional[Mapping[str, Any]] = None,
) -> AsyncIterator[Optional[sentry_sdk.tracing.Span]]:
    with Hub(Hub.current) as hub, hub.start_span(op="call", name=name) as child:
        if tags:
            for key, value in tags.items():
                child.set_tag(key, value)
        if data:
            for key, value in data.items():
                child.set_data(key, value)
        try:
            yield child
        except asyncio.CancelledError:
            child.set_status("cancelled")
            raise
        except Exception as exc:
            hub.capture_exception(error=exc)
            raise


@asynccontextmanager
async def trace_cm(
    name: str,
    tags: Optional[Mapping[str, str]] = None,
    data: Optional[Mapping[str, str]] = None,
) -> AsyncIterator[None]:
    async with sentry_trace_cm(name, tags=tags, data=data):
        yield


def trace(func: T) -> T:
    async def _tracer(*args: Any, **kwargs: Any) -> Any:
        name = func.__qualname__
        async with trace_cm(name):
            return await func(*args, **kwargs)

    @functools.wraps(func)
    async def tracer(*args: Any, **kwargs: Any) -> Any:
        # Create a task to wrap method call to avoid scope data leakage between calls.
        return await asyncio.create_task(_tracer(*args, **kwargs))

    return cast(T, tracer)


def new_trace(func: T) -> T:
    async def _tracer(*args: Any, **kwargs: Any) -> Any:
        name = func.__qualname__
        async with new_trace_cm(name):
            return await func(*args, **kwargs)

    @functools.wraps(func)
    async def tracer(*args: Any, **kwargs: Any) -> Any:
        # Create a task to wrap method call to avoid scope data leakage between calls.
        return await asyncio.create_task(_tracer(*args, **kwargs))

    return cast(T, tracer)


def new_sampled_trace(func: T) -> T:
    async def _tracer(*args: Any, **kwargs: Any) -> Any:
        name = func.__qualname__
        async with new_trace_cm(name, sampled=True):
            return await func(*args, **kwargs)

    @functools.wraps(func)
    async def tracer(*args: Any, **kwargs: Any) -> Any:
        # Create a task to wrap method call to avoid scope data leakage between calls.
        return await asyncio.create_task(_tracer(*args, **kwargs))

    return cast(T, tracer)


def notrace(func: T) -> T:
    @functools.wraps(func)
    async def tracer(*args: Any, **kwargs: Any) -> Any:
        with Hub.current.configure_scope() as scope:
            transaction = scope.transaction
            if transaction is not None:
                transaction.sampled = False
            return await func(*args, **kwargs)

    return cast(T, tracer)


def before_send_transaction(
    event: Event, hint: Hint, *, health_check_url_path: str
) -> Optional[Event]:
    url = URL(event["request"]["url"])  # type: ignore[arg-type]

    if url.path == health_check_url_path:
        return None

    return event


def _find_caller_version(stacklevel: int) -> str:
    caller = inspect.currentframe()
    assert caller is not None
    while stacklevel:
        caller = caller.f_back
        stacklevel -= 1
        assert caller is not None
    package, sep, tail = caller.f_globals["__package__"].partition(".")
    ver = version(package)
    return f"{package}@{ver}"


def setup_sentry(
    *,
    health_check_url_path: str = "/api/v1/ping",
    ignore_errors: Iterable[Union[type[BaseException], str]] = (),
) -> None:  # pragma: no cover
    config = EnvironConfigFactory().create_sentry()
    if config.dsn:
        LOGGER.info("Loaded Sentry config: %s", config)
    else:
        LOGGER.warning("Sentry DSN is not configured, skipping Sentry setup.")
        return
    ignore_errors = tuple(ignore_errors) + (
        asyncio.CancelledError,
        aiohttp.ServerConnectionError,
        ConnectionResetError,
    )
    sentry_sdk.init(
        dsn=str(config.dsn) or None,
        traces_sample_rate=config.sample_rate,
        integrations=[AioHttpIntegration(transaction_style="method_and_path_pattern")],
        ignore_errors=ignore_errors,
        before_send_transaction=functools.partial(
            before_send_transaction, health_check_url_path=health_check_url_path
        ),
        release=_find_caller_version(2),
        environment=config.cluster_name,
    )
    if config.app_name:
        sentry_sdk.set_tag("app", config.app_name)
    if config.cluster_name:
        sentry_sdk.set_tag("cluster", config.cluster_name)



import json
import logging
import os
import re
from typing import Any
from unittest import mock

import pytest
from dirty_equals import IsList, IsNow, IsPartialDict, IsPositiveInt, IsStr

from neuro_logging import AllowLessThanFilter, init_logging


@pytest.fixture(autouse=True)
def set_log_level() -> None:
    os.environ["LOG_LEVEL"] = "NOTSET"


def _log_all_messages() -> None:
    logging.debug("DebugMessage")
    logging.info("InfoMessage")
    logging.warning("WarningMessage")
    logging.error("ErrorMessage")
    logging.critical("CriticalMessage")


def test_default_config_format(capsys: Any) -> None:
    init_logging()
    logging.debug("DebugMessage")
    captured = capsys.readouterr()
    assert re.match(
        r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2},\d+ - root - DEBUG - DebugMessage",
        captured.out,
    )


def test_default_config_output(capsys: Any) -> None:
    init_logging()
    _log_all_messages()
    captured = capsys.readouterr()
    assert "DebugMessage" in captured.out
    assert "InfoMessage" in captured.out
    assert "WarningMessage" in captured.out
    assert "ErrorMessage" in captured.err
    assert "CriticalMessage" in captured.err


def test_health_checks_filtered(capsys: Any) -> None:
    init_logging()
    logging.getLogger("aiohttp.access").info("InfoMessage")
    logging.getLogger("aiohttp.access").info("GET /api/v1/ping")
    captured = capsys.readouterr()
    assert "InfoMessage" in captured.out
    assert "/api/v1/ping" not in captured.out


def test_health_checks_filtered__error(capsys: Any) -> None:
    init_logging()
    logging.getLogger("aiohttp.access").error("GET /api/v1/ping")
    captured = capsys.readouterr()
    assert "/api/v1/ping" in captured.err


def test_health_checks_filtered__custom_url_path(capsys: Any) -> None:
    init_logging(health_check_url_path="/health")
    logging.getLogger("aiohttp.access").info("GET /health")
    captured = capsys.readouterr()
    assert not captured.out


def test_allow_less_filter_usage() -> None:
    filter = AllowLessThanFilter(logging.INFO)
    record_info = logging.LogRecord("some", logging.INFO, "some", 12, "text", (), None)
    record_debug = logging.LogRecord(
        "some", logging.DEBUG, "some", 12, "text", (), None
    )
    assert filter.filter(record_info) is False
    assert filter.filter(record_debug) is True


def test_allow_less_filter_text_level_names() -> None:
    filter = AllowLessThanFilter("INFO")
    assert filter.level == logging.INFO

    with pytest.raises(ValueError):
        AllowLessThanFilter("unknown-level")


def test_existing_loggers_continue_work(capsys: Any) -> None:
    existing = logging.getLogger("existing")
    init_logging()

    existing.info("InfoMessage")
    existing.error("ErrorMessage")
    captured = capsys.readouterr()
    assert "InfoMessage" in captured.out
    assert "ErrorMessage" in captured.err


def test_json_logging_with_extra(capsys: Any, monkeypatch: Any) -> None:
    monkeypatch.delenv("PYTEST_VERSION")
    init_logging()
    logging.debug("msg", extra={"key": "first"})
    captured = capsys.readouterr()
    assert not captured.err
    msg = json.loads(captured.out)
    assert msg == IsPartialDict(
        {
            "args": [],
            "exc_info": None,
            "filename": "test_logging.py",
            "funcName": "test_json_logging_with_extra",
            "key": "first",
            "lineno": IsPositiveInt(),
            "logName": "root",
            "message": "msg",
            "module": "test_logging",
            "pathname": mock.ANY,
            "process": IsPositiveInt(),
            "processName": "MainProcess",
            "severity": "DEBUG",
            "stack_info": None,
            "thread": IsPositiveInt(),
            "threadName": "MainThread",
            "timestamp": mock.ANY,
        }
    )


def test_json_logging_with_args(capsys: Any, monkeypatch: Any) -> None:
    monkeypatch.delenv("PYTEST_VERSION")
    init_logging()
    logging.debug("%s msg", "arg")
    captured = capsys.readouterr()
    assert not captured.err
    msg = json.loads(captured.out)
    assert msg == IsPartialDict(
        {
            "args": ["arg"],
            "exc_info": None,
            "filename": "test_logging.py",
            "funcName": "test_json_logging_with_args",
            "lineno": IsPositiveInt(),
            "logName": "root",
            "message": "arg msg",
            "module": "test_logging",
            "pathname": mock.ANY,
            "process": IsPositiveInt(),
            "processName": "MainProcess",
            "severity": "DEBUG",
            "stack_info": None,
            "thread": IsPositiveInt(),
            "threadName": "MainThread",
            "timestamp": IsNow(tz="UTC", iso_string=True),
        }
    )


def test_json_logging_with_exc_info(capsys: Any, monkeypatch: Any) -> None:
    monkeypatch.delenv("PYTEST_VERSION")
    init_logging()
    try:
        1 / 0
    except ZeroDivisionError:
        logging.debug("%s msg", "arg", exc_info=True)
    captured = capsys.readouterr()
    assert not captured.err
    msg = json.loads(captured.out)
    assert msg == IsPartialDict(
        {
            "args": ["arg"],
            "exc_info": IsList(
                "ZeroDivisionError",
                "ZeroDivisionError: division by zero",
                length=3,
            ),
            "filename": "test_logging.py",
            "funcName": "test_json_logging_with_exc_info",
            "lineno": IsPositiveInt(),
            "logName": "root",
            "message": "arg msg",
            "module": "test_logging",
            "pathname": mock.ANY,
            "process": IsPositiveInt(),
            "processName": "MainProcess",
            "severity": "DEBUG",
            "stack_info": None,
            "thread": IsPositiveInt(),
            "threadName": "MainThread",
            "timestamp": IsNow(tz="UTC", iso_string=True),
        }
    )


def test_json_logging_with_stack_info(capsys: Any, monkeypatch: Any) -> None:
    monkeypatch.delenv("PYTEST_VERSION")
    init_logging()
    logging.debug("%s msg", "arg", stack_info=True)
    captured = capsys.readouterr()
    assert not captured.err
    msg = json.loads(captured.out)
    import pprint

    pprint.pprint(msg)
    assert msg == IsPartialDict(
        {
            "args": ["arg"],
            "exc_info": None,
            "filename": "test_logging.py",
            "funcName": "test_json_logging_with_stack_info",
            "lineno": IsPositiveInt(),
            "logName": "root",
            "message": "arg msg",
            "module": "test_logging",
            "pathname": mock.ANY,
            "process": IsPositiveInt(),
            "processName": "MainProcess",
            "severity": "DEBUG",
            "stack_info": IsStr(),
            "thread": IsPositiveInt(),
            "threadName": "MainThread",
            "timestamp": IsNow(tz="UTC", iso_string=True),
        }
    )
    assert msg["stack_info"].startswith("Stack (most recent call last):\n")



import asyncio
import re
import typing as t

import pytest
import sentry_sdk
from sentry_sdk.tracing import Span, Transaction

from neuro_logging.testing_utils import _get_test_version
from neuro_logging.trace import (
    before_send_transaction,
    new_sampled_trace,
    new_trace,
    notrace,
    trace,
    trace_cm,
)


@pytest.fixture()
def sentry_transaction() -> t.Iterator[None]:
    sentry_sdk.init(traces_sample_rate=1.0)
    hub = sentry_sdk.Hub.current
    with (
        sentry_sdk.Hub(hub) as hub,
        hub.start_transaction(
            Transaction(name="test", hub=hub, parent_sampled=True, sampled=True)
        ),
    ):
        yield


@pytest.mark.usefixtures("sentry_transaction")
@pytest.mark.xfail
async def test_sentry_trace() -> None:
    parent_span = sentry_sdk.Hub.current.scope.span

    @trace
    async def func() -> None:
        span = sentry_sdk.Hub.current.scope.span

        assert span
        assert parent_span != span
        assert span.op == "call"
        assert span.description == "test_sentry_trace.<locals>.func"

    await func()


@pytest.mark.usefixtures("sentry_transaction")
@pytest.mark.xfail
async def test_sentry_trace_cm_data() -> None:
    async with trace_cm(
        "test", tags={"test1": "val1", "test2": "val2"}, data={"data": "value"}
    ):

        span = sentry_sdk.Hub.current.scope.span

        assert span
        assert span._tags["test1"] == "val1"
        assert span._tags["test2"] == "val2"
        assert span._data["data"] == "value"


@pytest.mark.usefixtures("sentry_transaction")
@pytest.mark.xfail
async def test_sentry_trace_multiple_tasks() -> None:
    spans = []

    @trace
    async def func() -> None:
        await asyncio.sleep(0)
        span = sentry_sdk.Hub.current.scope.span

        assert span

        spans.append(span)

    await asyncio.gather(func(), func())

    span1, span2 = spans

    assert span1.span_id != span2.span_id


@pytest.mark.xfail
async def test_sentry_new_trace() -> None:
    @new_trace
    async def func() -> None:
        span = sentry_sdk.Hub.current.scope.span

        assert isinstance(span, Transaction)
        assert span.name == "test_sentry_new_trace.<locals>.func"

    sentry_sdk.init(traces_sample_rate=1.0)

    await func()


@pytest.mark.xfail
async def test_sentry_new_trace_multiple_tasks() -> None:
    sentry_sdk.init(traces_sample_rate=1.0)
    spans: list[t.Optional[Span]] = []

    @new_trace
    async def func() -> None:
        await asyncio.sleep(0)
        span = sentry_sdk.Hub.current.scope.span

        spans.append(span)

    await asyncio.gather(func(), func())

    span1, span2 = spans

    assert span1
    assert span2
    assert span1.trace_id != span2.trace_id


@pytest.mark.xfail
async def test_sentry_new_sampled_trace() -> None:
    @new_sampled_trace
    async def func() -> None:
        span = sentry_sdk.Hub.current.scope.span

        assert isinstance(span, Transaction)
        assert span.name == "test_sentry_new_sampled_trace.<locals>.func"
        assert span.sampled is True

    sentry_sdk.init(traces_sample_rate=1.0)

    await func()


@pytest.mark.usefixtures("sentry_transaction")
async def test_sentry_notrace() -> None:
    @notrace
    async def func() -> None:
        assert not sentry_sdk.Hub.current.scope.transaction.sampled

    assert sentry_sdk.Hub.current.scope.transaction.sampled

    await func()


def test_find_caller_version() -> None:
    version = _get_test_version()
    assert re.match(r"^neuro_logging@\d+[.]\d+", version)


def test_sentry_before_send_transaction() -> None:
    event = before_send_transaction(
        {"request": {"url": "http://127.0.0.1/api/v1/ping"}},
        {},
        health_check_url_path="/api/v1/ping",
    )
    assert event is None

    event = before_send_transaction(
        {"request": {"url": "http://127.0.0.1/api/v1/jobs"}},
        {},
        health_check_url_path="/api/v1/ping",
    )
    assert event is not None
